

def get_printables(
                   axes: str,
                   shapedict: dict,
                   scaledict: dict,
                   unitdict: dict
                   ):
    dimensions = [dim for dim in axes if dim in scaledict.keys()]

    rows = [("Dimension", "Size (pixels)", "Scale", "Unit")]
    for i, dim in enumerate(dimensions):
        # size = shape[i] if i < len(shape) else ''
        size = shapedict.get(dim, '')
        scale = scaledict.get(dim, '')
        unit = unitdict.get(dim, '')
        rows.append((dim, str(size), str(scale), unit))

    col_widths = [max(len(str(row[i])) for row in rows) for i in range(4)]

    printables = []

    # ANSI escape codes for bold
    BOLD = '\033[1m'
    RESET = '\033[0m'

    for idx, row in enumerate(rows):
        line = "  ".join(str(cell).ljust(col_widths[i]) for i, cell in enumerate(row))
        if idx == 0:
            line = BOLD + line + RESET
        printables.append(line)

    return printables

def print_printable(printable):
    for item in printable:
        print(item)

# def show_pixel_meta(input_path):
#     # input_path = f"/home/oezdemir/PycharmProjects/dask_env1/data/tifflist"
#     base = BridgeBase(input_path)
#     base.read_dataset(True)
#     base.digest()
#     base.compute_pixel_metadata()
#     ###
#     printables = {}
#     for path, vmeta in base.pixel_metadata.vmetaset.items():
#         shape = vmeta.shape
#         scaledict = vmeta.scaledict
#         unitdict = vmeta.unitdict
#         printable = get_printables(shape,scaledict,unitdict)
#         printables[path] = printable
#     for path, printable in printables.items():
#         print('---------')
#         print(f"")
#         print(f"Metadata for '{path}':")
#         print_printable(printable)




default_axes = 'tczyx'

unit_map = {
    't': 'second',
    'c': 'Channel',
    'z': 'micrometer',
    'y': 'micrometer',
    'x': 'micrometer'
}



scale_factor_map = {
    't': 1,
    'c': 1,
    'z': 1,
    'y': 2,
    'x': 2
}

scale_map = {
    't': 1,
    'c': 1,
    'z': 1,
    'y': 1,
    'x': 1
}

type_map = {
    't': 'time',
    'c': 'channel',
    'z': 'space',
    'y': 'space',
    'x': 'space'
}

datasets_archetype = {
                '0': {
                    'array': None,
                    'axis_order': None,
                    'scale': None,
                    'unit_list': None,
                    'chunks': None,
                    'dtype': None
                },
                '1': {
                    'array': None,
                    'axis_order': None,
                    'scale': None,
                    'unit_list': None,
                    'chunks': None,
                    'dtype': None
                }
        }

NotMultiscalesException = Exception('Not applicable as this group is not a single image or label.')
NotImageException = Exception('Not applicable as this group is not an image.')
AxisNotFoundException = Exception('Axis information not provided in the metadata.')

TempDir = '/tmp/OME-Zarr'

refpath = '0'

rechunker_maxmem = '1G'

array_meta_keys = ['chunks', 'shape', 'compressor', 'dtype', 'dimension_separator']




from eubi_bridge import ebridge
from eubi_bridge import ebridge_base
from eubi_bridge import fileset_io
from eubi_bridge.base import scale, writers
from eubi_bridge.ngff import defaults, multiscales
from eubi_bridge.utils import convenience, dask_client_plugins

__all__ = [
    'ebridge',
    'ebridge_base',
    'fileset_io',
    'scale',
    'writers',
    'defaults',
    'multiscales',
    'convenience',
    'dask_client_plugins',
]



import copy
import re
import os
from typing import Dict, Iterable, List, Union

import dask.array as da
import numpy as np
from dask import delayed
from natsort import natsorted

from eubi_bridge.base.data_manager import ArrayManager, ChannelIterator

transpose_list = lambda l: list(map(list, zip(*l)))
get_numerics = lambda string: list(re.findall(r'\d+', string))
get_alpha = lambda string: ''.join([i for i in string if not i.isnumeric()])


def get_matches(pattern, strings, return_non_matches=False):
    """
    Search for regex pattern matches in a list of strings.

    Args:
        pattern (str): Regular expression pattern to search for.
        strings (list): List of strings to search within.
        return_non_matches (bool): If True, returns all results including None for non-matches.
                                 If False, returns only successful matches.

    Returns:
        list: List of match objects (or None for non-matches if return_non_matches is True).
    """
    matches = [re.search(pattern, string) for string in strings]
    if return_non_matches:
        return matches
    return [match for match in matches if match is not None]


def split_by_match(filepaths, *args):
    """
    Group filepaths based on matching patterns.

    Args:
        filepaths (list): List of file paths to search in.
        *args: Variable number of patterns to search for in filepaths.

    Returns:
        dict: Dictionary with patterns as keys and lists of matching filepaths as values.
    """
    ret = dict().fromkeys(args)
    for key in args:
        matches = get_matches(key, filepaths)
        ret[key] = matches
    return ret


def find_match_and_numeric(filepaths, *args):
    """
    Find matches in filepaths and group them by their numeric suffixes.

    Args:
        filepaths (list): List of file paths to search in.
        *args: Variable number of patterns to search for in filepaths.

    Returns:
        dict: Dictionary where keys are matched patterns and values are lists of match objects.
    """
    ret = {}
    for key in args:
        matches = get_matches(key, filepaths)
        for match in matches:
            span = match.string[match.start():match.end()]
            if span not in ret.keys():
                ret[span] = [match]
            else:
                ret[span].append(match)
    return ret


def concatenate_shapes_along_axis(shapes: Iterable,
                                  axis: int
                                  ) -> list:
    """
    Concatenate shapes along a specified axis.

    Args:
        shapes (Iterable): Iterable of shape tuples to concatenate.
        axis (int): Axis along which to concatenate shapes.

    Returns:
        list: New shape after concatenation along the specified axis.

    Raises:
        ValueError: If dimensions other than the concatenation axis don't match.
    """
    reference_shape = shapes[0]
    concatenated_shape = [num for num in reference_shape]
    for shape in shapes[1:]:
        for idx, size in enumerate(shape):
            if idx == axis:
                concatenated_shape[idx] += size
            else:
                assert size == reference_shape[idx], ValueError(
                    "For concatenation to succeed, all dimensions except the dimension of concatenation must match.")
    return concatenated_shape


def accumulate_slices_along_axis(shapes: Iterable,
                                 axis: int,
                                 slices: Union[tuple, list] = None
                                 ) -> list:
    """
    Calculate accumulated slices for concatenation along a specified axis.

    Args:
        shapes (Iterable): Iterable of shape tuples to be concatenated.
        axis (int): Axis along which to calculate slices.
        slices (Union[tuple, list], optional): Initial slices. If None, creates full slices.

    Returns:
        list: List of slice tuples for each input shape.
    """
    reference_shape = shapes[0]
    if slices is None:
        slices = [[slice(None, None) for _ in reference_shape] for _ in shapes]
    assert (len(shapes) == len(slices))
    sizes_per_axis = [shape[axis] for shape in shapes]
    cummulative_sizes = [0] + np.cumsum(sizes_per_axis).tolist()
    slice_tuples = [(cummulative_sizes[idx], cummulative_sizes[idx + 1]) for idx in range(len(sizes_per_axis))]
    for idx, tup in enumerate(slice_tuples):
        slc = slices[idx]
        slclist = list(slc)
        slclist[axis] = slice(*tup)
        slices[idx] = tuple(slclist)
    return slices


def reduce_paths_flexible(paths: Iterable[str],
                          dimension_tag: Union[str, tuple, list],
                          replace_with: str = 'set') -> str:
    """
    Reduces a list of similar paths by merging over the specified dimension.

    - If `dimension_tag` is a string (e.g., 'T' or 'Channel'), it's assumed to be followed by digits;
      the digits are replaced with `replace_with`.
    - If `dimension_tag` is a tuple/list (e.g., ('blue', 'red')), those are treated as categorical tokens
      and replaced with their joined value plus `replace_with`.
    """
    paths = list(paths)
    if not paths:
        return ""

    if isinstance(dimension_tag, str):
        # Match like 'T0001', 'Channel2', etc.
        pattern = re.compile(rf'({re.escape(dimension_tag)})(\d+)')
        def replace_tag(path):
            return pattern.sub(lambda m: m.group(1) + replace_with, path)

    elif isinstance(dimension_tag, (tuple, list)):
        # Categorical case: match only if surrounded by boundaries like /, _, -, ., or start/end of string
        unique_vals = sorted(set(dimension_tag))
        joined_val = ''.join(unique_vals) + replace_with

        # Example: match (^|/|_|-|.)blue(?=$|/|_|-|.)
        pattern = re.compile(
            rf'(?:(?<=^)|(?<=[/_\-.]))({"|".join(map(re.escape, unique_vals))})(?=$|[\/_\-.])'
        )

        def replace_tag(path):
            return pattern.sub(joined_val, path)

    else:
        raise ValueError("dimension_tag must be a string or a tuple/list of strings")

    # Apply replacement
    replaced_paths = [replace_tag(p) for p in paths]

    # Now combine all paths token-wise
    tokenized = [re.split(r'([/_\-.])', p) for p in replaced_paths]
    merged_tokens = []
    for tokens in zip(*tokenized):
        uniq = list(dict.fromkeys(tokens))  # preserve order
        merged_tokens.append(''.join(uniq))

    return ''.join(merged_tokens)


class FileSet:
    """
    A class to manage file paths and their shapes for multi-dimensional data.

    This class handles file paths and their corresponding array data, supporting operations
    like concatenation along specified axes. It's designed to work with up to 5 dimensions (t, c, z, y, x).
    """

    AXIS_DICT = {
        0: 't',
        1: 'c',
        2: 'z',
        3: 'y',
        4: 'x'
    }


    def __init__(self,
                 filepaths: Iterable[str],
                 shapes: Iterable[tuple | list] = None,
                 axis_tag0: Union[str, tuple] = None,
                 axis_tag1: Union[str, tuple] = None,
                 axis_tag2: Union[str, tuple] = None,
                 axis_tag3: Union[str, tuple] = None,
                 axis_tag4: Union[str, tuple] = None,
                 arrays: Iterable[da.Array] = None):
        """
        Initialize the FileSet class.

        Args:
            filepaths: The file paths of the arrays.
            shapes: The shapes of the arrays. Required if arrays is not provided.
            axis_tag0-4: Tags for each axis (t, c, z, y, x).
            arrays: The arrays. If provided, shapes can be None.
        """
        if shapes is None and arrays is None:
            raise ValueError("Either shapes or arrays must be provided.")

        if arrays is not None:
            self.array_dict = dict(zip(filepaths, arrays))
            shapes = [arr.shape for arr in arrays]
        else:
            self.array_dict = None

        self.shape_dict = dict(zip(filepaths, shapes))
        self.axis_tags = [axis_tag0, axis_tag1, axis_tag2, axis_tag3, axis_tag4]

        # Initialize dimension tags and specified axes
        self.dimension_tags = []
        self.specified_axes = []
        for axis, tag in enumerate(self.axis_tags):
            if tag is not None:
                self.dimension_tags.append(tag)
                self.specified_axes.append(axis)

        self.group = {'': list(filepaths)}
        self.slice_dict = {
            path: tuple(slice(0, size) for size in shape)
            for path, shape in self.shape_dict.items()
        }
        self.path_dict = dict(zip(filepaths, filepaths))


    def get_numerics_per_dimension_tag(self,
                                       dimension_tag: str
                                       ) -> List[str]:
        """
        Extract numeric values from filepaths for a given dimension tag.

        Args:
            dimension_tag (str): The dimension tag to extract numerics for
                (e.g., 't' for time).

        Returns:
            list: List of numeric strings extracted from the filepaths.

        Example:
            >>> f = FileSet(['file_t0001_channel1.ome.tif', 'file_t0002_channel2.ome.tif'])
            >>> f.get_numerics_per_dimension_tag('t')
            ['0001', '0002']
        """
        filepaths = list(self.group.values())[0]
        matches = get_matches(f'{dimension_tag}\d+', filepaths)
        spans = [match.string[match.start():match.end()] for match in matches]
        numerics = [get_numerics(span)[0] for span in spans]
        # TODO: add an incrementality validator
        return numerics


    def _csplit_by(self, tup: tuple) -> dict:
        """
        Split the filepaths in the group by the given dimension tags.

        Args:
            tup (tuple): A tuple of dimension tags to split by.

        Returns:
            dict: The split group as a dictionary.
        """
        group = copy.deepcopy(self.group)
        for key, filepaths in group.items():
            # Initialize a dictionary to store the split filepaths
            alpha_dict = {key: [] for key in tup}
            for tag in tup:
                # Get matches for the current dimension tag
                matches = get_matches(f'{tag}', filepaths)
                # Extract the matched spans
                spans = [match.string[match.start():match.end()] for match in matches]
                # Extract the matched filepaths
                matched_paths = [match.string for match in matches]
                # Create a copy of the spans
                alpha = copy.deepcopy(spans)
                # Get the unique categories for the current dimension tag
                alpha_categories = np.unique(alpha).tolist()
                # Check that there is only one category
                assert len(alpha_categories) == 1, f"Number of categories is not 1: {alpha_categories}"
                # Get the category tag
                alpha_tag = alpha_categories[0]
                # Store the matched filepaths in the alpha dictionary
                alpha_dict[alpha_tag] = matched_paths
            # Update the group with the split filepaths
            group = alpha_dict
        return group


    def _split_by(self, *args):
        """
        Split the filepaths in the group by the given dimension tags.

        Args:
            *args (str): The dimension tags to split by.

        Returns:
            dict: The split group as a dictionary.
        """
        group = copy.deepcopy(self.group)
        for dim in args:
            if dim not in self.dimension_tags:
                raise ValueError(f"The dimension '{dim}' is not among the given dimension_tags.")
            # If the dimension tag is a tuple or list, split by all of them
            if isinstance(dim, (tuple, list)):
                group = self._csplit_by(dim)
            else:
                numeric_dict = {}
                for key, filepaths in group.items():
                    matches = get_matches(f'{dim}\d+', filepaths)
                    spans = [match.string[match.start():match.end()] for match in matches]
                    spans = [span.replace(dim, '') for span in spans]  ### remove search term from the spans
                    numerics = [get_numerics(span)[0] for span in spans]
                    numeric_categories = np.unique(numerics).tolist()
                    for idx, num in enumerate(numerics):
                        for i, category in enumerate(numeric_categories):
                            if num == category:
                                if key != '':
                                    tag_key = ''.join([key, '-', dim, num])
                                else:
                                    tag_key = ''.join([dim, num])
                                if not tag_key in numeric_dict:
                                    numeric_dict[tag_key] = []
                                numeric_dict[tag_key].append(filepaths[idx])
                group = numeric_dict
        return group

    def concatenate_along(self, axis: int) -> dict:
        """
        Concatenate arrays along the specified axis.

        Args:
            axis: The axis along which to concatenate the arrays.

        Returns:
            dict: The grouped file paths after concatenation.

        Raises:
            ValueError: If the axis is not among the given dimension tags.
        """
        dimension_tag = self.axis_tags[axis]
        if dimension_tag not in self.dimension_tags:
            raise ValueError(f"The dimension '{dimension_tag}' is not among the given dimension_tags.")

        # Split the group by all dimension tags except the one specified by the axis
        to_split = [tag for tag in self.dimension_tags if tag != dimension_tag]
        group = self._split_by(*to_split)

        for key, paths in group.items():
            # Sort paths naturally
            sorted_paths = natsorted(paths)

            # Get slices and shapes for each path
            group_slices = [self.slice_dict[path] for path in sorted_paths]
            group_shapes = [self.shape_dict[path] for path in sorted_paths]
            group_reduced_paths = [self.path_dict[path] for path in sorted_paths]

            # Calculate new slices and shape after concatenation
            new_slices = accumulate_slices_along_axis(group_shapes, axis, group_slices)
            new_shape = concatenate_shapes_along_axis(group_shapes, axis)

            # Update paths with concatenated version
            new_reduced_path = reduce_paths_flexible(
                group_reduced_paths,
                dimension_tag,
                replace_with=f'_{self.AXIS_DICT[axis]}set'
            )
            new_reduced_paths = [new_reduced_path] * len(group_reduced_paths)

            # If arrays are present, concatenate them
            if self.array_dict is not None:
                group_arrays = [self.array_dict[path] for path in sorted_paths]
                new_array = da.concatenate(group_arrays, axis=axis)

            # Update dictionaries with new values
            for path, slc, reduced_path in zip(sorted_paths, new_slices, new_reduced_paths):
                self.slice_dict[path] = slc
                self.shape_dict[path] = new_shape
                self.path_dict[path] = reduced_path
                if self.array_dict is not None:
                    self.array_dict[path] = new_array

        return group

    def get_concatenated_arrays(self) -> Dict[str, tuple]:
        """
        Get a dictionary of concatenated arrays with their metadata.

        Returns:
            dict: A dictionary where keys are input paths and values are tuples of
                (updated_path, array_data).
        """
        # Get unique paths and their corresponding keys
        unique_paths = []
        unique_input_paths = []
        unique_ids = []

        # Process paths in natural sort order
        for key in natsorted(self.path_dict):
            path = self.path_dict[key]
            if path not in unique_paths:
                unique_input_paths.append(key)
                unique_paths.append(path)
                unique_ids.append(key)

        # Get arrays for unique paths
        unique_arrays = [self.array_dict[path] for path in unique_ids]

        # Create result dictionary
        return {
            key: (path, arr)
            for key, path, arr in zip(unique_input_paths, unique_paths, unique_arrays)
        }


@delayed
def build_managers_dict(**delayed_managers):
    return delayed_managers


class BatchFile:
    def __init__(self,
                 filepaths: Iterable[str],
                 shapes: Iterable[tuple | list] = None,
                 axis_tag0: Union[str, tuple] = None,
                 axis_tag1: Union[str, tuple] = None,
                 axis_tag2: Union[str, tuple] = None,
                 axis_tag3: Union[str, tuple] = None,
                 axis_tag4: Union[str, tuple] = None,
                 arrays: Iterable[da.Array] = None,
                 ):

        self.fileset = FileSet(filepaths,
                          shapes=shapes,
                          axis_tag0=axis_tag0,
                          axis_tag1=axis_tag1,
                          axis_tag2=axis_tag2,
                          axis_tag3=axis_tag3,
                          axis_tag4=axis_tag4,
                          arrays=arrays)
        self.managers = None


    def split_channel_groups(self):

        fileset = self.fileset
        sub_filesets = {}
        axis_tags = copy.deepcopy(fileset.axis_tags)

        if all([item is None for item in axis_tags]):
            groups = copy.deepcopy(fileset.path_dict)
            for key, value in groups.items():
                groups[key] = [value]
        elif fileset.axis_tags[1] is None:
            groups = copy.deepcopy(fileset.group)
        else:
            axis_tags[1] = None
            groups = fileset._split_by(fileset.axis_tags[1])
        return groups


    def read_metadata(self,
                      series: int = None,
                      metadata_reader: str = 'bfio',
                      **kwargs
                      ):
        grs = self.split_channel_groups()
        self.sample_paths = natsorted([grs[grname][0]
                                        for grname in grs])
        managers = {path: delayed(ArrayManager)(path,
                                            series = series,
                                            metadata_reader = metadata_reader,
                                            **kwargs
                                            ) for
                         path in self.sample_paths
                         }

        self.managers = build_managers_dict(**managers).compute()
        return self.managers

    def batch_concatenate(self,
                          axes: Iterable[int] = []
                          ):

        for axis in axes:
            self.fileset.concatenate_along(axis)

        if self.managers is None:
            self.managers = self.read_metadata()
        channels_ = {}
        for path, manager in self.managers.items():
            manager._ensure_correct_channels()
            manager.fix_bad_channels()
            channels_[manager.path] = manager.channels

        channelslist = []
        channels = {}
        if 1 in axes:
            for chkey in natsorted(channels_.keys()):
                channelslist.extend(channels_[chkey])
            for chkey in natsorted(channels_.keys()):
                channels[chkey] = channelslist
        else:
            channels = channels_

        self.channels = channels


    def _update_nonunique_channel_colors(self,
                                         channels
                                         ):
        colors = [channel['color'] for channel in channels]
        if len(set(colors)) < len(colors):
            chn = ChannelIterator(num_channels=len(colors))
            for channel, _channel in zip(channels, chn._channels):
                channel['color'] = _channel['color']
        return channels


    def get_output_dicts(self,
                         root_path,
                         path_separator: str = '-',
                         ):
        fileset = self.fileset
        root_path_ = os.path.normpath(root_path).split(os.sep)
        root_path_top = []
        for item in root_path_:
            if '*' in item:
                break
            root_path_top.append(item)

        if os.name == 'nt':
            # Use os.path.splitdrive to handle any drive letter
            drive, _ = os.path.splitdrive(root_path)
            root_path = os.path.join(drive + os.sep, *root_path_top)
        else:
            root_path = os.path.join(os.sep, *root_path_top)

        arrays_ = fileset.get_concatenated_arrays()  ### ! Careful here


        arrays, channels, sample_paths, managers = {}, {}, {}, {}

        for key, vals in arrays_.items():
            (updated_key, arr,
             ) = vals
            new_key = os.path.relpath(updated_key, root_path)
            new_key = os.path.splitext(new_key)[0]
            new_key = new_key.replace(os.sep, path_separator)
            arrays[new_key] = arrays_[key][1]
            sample_paths[new_key] = key

            ### Update colors if they are not unique
            self.channels[key] = self._update_nonunique_channel_colors(self.channels[key])
            ###

            channels[new_key] = self.channels[key]
            managers[new_key] = self.managers[key]
            managers[new_key]._channels = self.channels[key]


        return (arrays,
                sample_paths,
                managers)



import numpy as np
import warnings
import dask.array as da
import os
from typing import Optional, Dict, Any, Tuple, Union
import warnings
import dask.array as da


def read_czi(
    input_path: str,
    as_mosaic: bool = False,
    view_index: int = 0,
    phase_index: int = 0,
    illumination_index: int = 0,
    scene_index: int = 0,
    rotation_index: int = 0,
    mosaic_tile_index: int = 0,
    sample_index: int = 0
) -> da.Array:
    """
    Read a CZI (Zeiss microscopy) file with specified dimension indices.

    Args:
        input_path: Path to the CZI file.
        as_mosaic: Whether to read as a mosaic (tiled) image.
        view_index: Index for the view dimension (v).
        phase_index: Index for the phase dimension (h).
        illumination_index: Index for the illumination dimension (i).
        scene_index: Index for the scene dimension (s).
        rotation_index: Index for the rotation dimension (r).
        mosaic_tile_index: Index for the mosaic tile dimension (m).
        sample_index: Index for the sample dimension (a).

    Returns:
        dask.array.Array: The image data with dimension order TCZYX.

    Raises:
        FileNotFoundError: If the input file doesn't exist.
        ValueError: If invalid dimension indices are provided.
    """
    if not os.path.isfile(input_path):
        raise FileNotFoundError(f"File not found: {input_path}")

    # Import the appropriate reader
    if as_mosaic:
        from bioio_czi.pylibczirw_reader.reader import Reader
    else:
        from bioio_czi.aicspylibczi_reader.reader import Reader

    # Initialize reader and get image metadata
    try:
        img = Reader(input_path)
    except Exception as e:
        raise RuntimeError(f"Failed to read CZI file: {str(e)}") from e

    # Handle scene selection
    if scene_index != 0:
        if not 0 <= scene_index < len(img.scenes):
            raise ValueError(
                f"Scene index {scene_index} is out of range. "
                f"Must be between 0 and {len(img.scenes) - 1}."
            )
        img.set_scene(scene_index)

    # Process non-standard dimensions
    nonstandard_dims = [
        dim.upper() for dim in img.standard_metadata.dimensions_present
        if dim.upper() not in {"X", "Y", "C", "T", "Z"}
    ]

    # Handle mosaic-specific logic
    if as_mosaic:
        if 'M' in nonstandard_dims:
            nonstandard_dims.remove('M')
        else:
            warnings.warn(
                f"Mosaic tile dimension not found in {input_path}. "
                "Ignoring 'as_mosaic' parameter."
            )
            as_mosaic = False
        if mosaic_tile_index != 0:
            warnings.warn(
                "Mosaic tile index is ignored when reading the entire mosaic. "
                "Set as_mosaic=False to read specific tiles."
            )

    # Map dimension indices
    czi_dim_map = {
        'V': view_index,
        'H': phase_index,
        'I': illumination_index,
        'S': scene_index,
        'R': rotation_index,
        'M': mosaic_tile_index,
        'A': sample_index
    }

    # Create index map for non-standard dimensions
    index_map = {
        dim: czi_dim_map[dim]
        for dim in nonstandard_dims
        if dim in czi_dim_map
    }

    # Read the image data
    try:
        return img.get_image_dask_data(dimension_order_out='TCZYX', **index_map)
    except Exception as e:
        raise RuntimeError(f"Failed to read image data: {str(e)}") from e








# from watchdog.observers import Observer
import os, asyncio
import signal
# from watchdog.observers.polling import PollingObserver
# from watchdog.events import FileSystemEventHandler, FileSystemEvent
from distributed.diagnostics.plugin import SchedulerPlugin
from distributed.diagnostics.plugin import NannyPlugin
from distributed.diagnostics.plugin import SchedulerPlugin
from distributed import get_client, Client


class ShutdownOnWorkerRemovalPlugin(SchedulerPlugin):
    async def remove_worker(self, scheduler, worker, **kwargs):
        for worker in scheduler.workers.values():
            await scheduler.remove_worker(worker.address, stimulus_id="shutdown")
            print(f"Worker {worker} has been removed. Shutting down process.")
        await asyncio.sleep(5)
        await scheduler.close()
        # os.kill(os.getpid(), signal.SIGTERM)

class ForcefulShutdownPlugin(SchedulerPlugin):
    def start(self, scheduler):
        for idx, worker in dask_scheduler.workers.items():
            print(f"Worker shutting down: {idx}")
            scheduler.remove_worker(worker.address, stimulus_id="shutdown")



# class RecursiveWriteActivityHandler(FileSystemEventHandler):
#     def __init__(self):
#         self.last_save_moment = None
#     def on_created(self, event: FileSystemEvent) -> None:
#         self.last_save_moment = time.time()
#         self.event_details = [event.src_path, event.event_type]
#     # def on_modified(self, event: FileSystemEvent) -> None:
#     #     self.last_save_moment = time.time()
#     #     self.event_details = [event.src_path, event.event_type]
#     # def on_moved(self, event: FileSystemEvent) -> None:
#     #     self.last_save_moment = time.time()
#     #     self.event_details = [event.src_path, event.event_type]

# async def shutdown_callback(dask_scheduler,
#                             target,
#                             timeout = 60,
#                             verbose = False
#                             ):
#
#     observer = PollingObserver()
#     handler = RecursiveWriteActivityHandler()
#     observer.schedule(handler, path = target, recursive = True)
#
#     observer.start()
#     try:
#         while True:
#             await asyncio.sleep(5)
#             if handler.last_save_moment is None:
#                 # print(f"None")
#                 pass
#             else:
#                 if verbose:
#                     print(handler.event_details)
#                 interval = time.time() - handler.last_save_moment
#                 if  interval > timeout:
#                     print(f"{interval} seconds passed in the target folder without action.")
#                     print(f"Killing the workers.")
#                     for worker in dask_scheduler.workers.values():
#                         await dask_scheduler.remove_worker(worker.address, stimulus_id="shutdown")
#                     print(f"Closing the scheduler.")
#                     await dask_scheduler.close()
#                     observer.stop()
#                     break
#     finally:
#         observer.stop()
#         observer.join()


class ShutdownPlugin(SchedulerPlugin):
    def __init__(self, output_folder, timeout = 60, verbose = False):
        self.output_folder = output_folder
        self.timeout = timeout
        self.verbose = verbose

    def start(self, scheduler):
        asyncio.create_task(shutdown_callback(scheduler, self.output_folder, self.timeout, self.verbose))


import zarr, dataclasses
from pathlib import Path
import numpy as np, zarr
import dask.array as da
from typing import Callable


def simple_downscale(
                     darr,
                     scale_factor: (tuple, list, np.ndarray) = None,
                     backend = 'numpy' # placeholder
                     ):
    """
    Downscale a Dask array along each dimension by given scale factors.

    Parameters:
    arr (dask.array): The input n-dimensional Dask array.
    scale_factors (tuple): The downsampling factors for each dimension.

    Returns:
    dask.array: The downscaled Dask array.
    """
    if len(scale_factor) != darr.ndim:
        raise ValueError("scale_factors must have the same length as the array's number of dimensions")
    slices = tuple(slice(None, None, int(scale)) for scale in scale_factor)
    downscaled_arr = darr[slices]
    return downscaled_arr

def mean_downscale(arr: da.Array,
                   scale_factor: (tuple, list, np.ndarray) = None
                   ):
    if len(scale_factor) != arr.ndim:
        raise ValueError("scale_factors must have the same length as the array's number of dimensions")
    axes = dict({idx: factor for idx, factor in enumerate(scale_factor)})
    downscaled_arr = da.coarsen(da.mean, arr,
                                axes = axes, trim_excess = True).astype(arr.dtype)
    return downscaled_arr

def median_downscale(arr: da.Array,
                   scale_factor: (tuple, list, np.ndarray) = None
                   ):
    if len(scale_factor) != arr.ndim:
        raise ValueError("scale_factors must have the same length as the array's number of dimensions")
    axes = dict({idx: factor for idx, factor in enumerate(scale_factor)})
    downscaled_arr = da.coarsen(da.median, arr,
                                axes = axes, trim_excess = True).astype(arr.dtype)
    return downscaled_arr

@dataclasses.dataclass
class DownscaleManager:
    base_shape: (list, tuple)
    scale_factor: (list, tuple)
    n_layers: (list, tuple)
    scale: (list, tuple) = None

    def __post_init__(self):
        ndim = len(self.base_shape)
        assert len(self.scale_factor) == ndim

    @property
    def _scale_ids(self):
        return np.arange(self.n_layers).reshape(-1, 1)

    @property
    def _theoretical_scale_factors(self):
        return np.power(self.scale_factor, self._scale_ids)

    @property
    def output_shapes(self):
        shapes = np.floor_divide(self.base_shape, self._theoretical_scale_factors)
        shapes[shapes == 0] = 1
        return shapes

    @property
    def scale_factors(self):
        return np.true_divide(self.output_shapes[0], self.output_shapes)

    @property
    def scales(self):
        return np.multiply(self.scale, self.scale_factors)


@dataclasses.dataclass
class Downscaler:
    array: da.Array
    scale_factor: (list, tuple)
    n_layers: int
    scale: (list, tuple) = None
    output_chunks: (list, tuple) = None
    backend: str = 'numpy'
    downscale_method: str = 'simple'

    def __post_init__(self):
        self.param_names = ['array', 'scale_factor', 'n_layers', 'scale', 'output_chunks', 'backend', 'downscale_method']
        self.update()

    def get_method(self):
        if self.downscale_method == 'simple':
            method = simple_downscale
        elif self.downscale_method == "mean":
            method = mean_downscale
        elif self.downscale_method == "median":
            method = mean_downscale
        else:
            raise NotImplementedError(f"Currently, only 'simple', 'mean' and 'median' methods are implemented.")
        return method

    def run(self):
        self.method = self.get_method()
        assert isinstance(self.array, da.Array)
        self.dm = DownscaleManager(self.array.shape,
                                   self.scale_factor,
                                   self.n_layers,
                                   self.scale
                                   )
        if self.output_chunks is None:
            self.output_chunks = [self.array.chunksize] * self.n_layers

        downscaled = []
        for idx, (scale_factor, chunks) in enumerate(zip(self.dm.scale_factors, self.output_chunks)):
            if idx == 0:
                downscaled.append(self.array)
            else:
                factor = tuple(int(x) for x in scale_factor)
                res1 = self.method(self.array, scale_factor = factor)
                downscaled.append(res1)
        self.downscaled_arrays = {str(i): arr for i, arr in enumerate(downscaled)}
        return self

    def update(self, **kwargs):
        for key, value in kwargs.items():
            if key in self.param_names:
                self.__setattr__(key, value)
            else:
                warnings.warn(f"The given parameter name '{key}' is not valid, ignoring it..")
        self.run()
        return self



# -*- coding: utf-8 -*-
"""
@author: bugra
"""

import setuptools
import os

def get_requirements():
    """Get requirements from requirements.txt or return default requirements."""
    requirements = [
        "aicspylibczi>=0.0.0",
        "asciitree>=0.3.3",
        "bfio>=0.0.0",
        "bioformats_jar>=0.0.0",
        "bioio-base>=0.0.0",
        "bioio-bioformats==1.1.0",
        "bioio-czi==2.1.0",
        "bioio-imageio==1.1.0",
        "bioio-lif==1.1.0",
        "bioio-nd2==1.1.0",
        "bioio-ome-tiff-fork-by-bugra==0.0.1b2",
        "bioio-tifffile-fork-by-bugra>=0.0.1b2",
        "cmake==4.0.2",
        "dask>=2024.12.1",
        "dask-jobqueue>=0.0.0",
        "distributed>=2024.12.1",
        "elementpath==5.0.1",
        "fasteners==0.19",
        "fire>=0.0.0",
        "imageio==2.27.0",
        "imageio-ffmpeg==0.6.0",
        "natsort>=0.0.0",
        "nd2>=0.0.0",
        "numpy>=0.0.0",
        # "openjdk==8.*",
        "pydantic>=0.0.0",
        "pylibczirw>=0.0.0",
        "readlif==0.6.5",
        "s3fs>=0.0.0",
        "scipy>=1.8",
        "tensorstore>=0.0.0",
        "tifffile>=2025.5.21",
        "validators==0.35.0",
        "xarray>=0.0.0",
        "xmlschema>=0.0.0",
        "xmltodict==0.14.2",
        "zarr>=3.0",
        "zstandard>=0.0.0"
    ]

    # Optionally still try to read from requirements.txt if it exists
    # if os.path.exists('../requirements.txt'):
    #     with open('../requirements.txt', encoding='utf-8') as f:
    #         requirements = [
    #             line.strip() for line in f
    #             if line.strip() and not line.startswith('#')
    #         ]
    return requirements

def readme():
    """Read the README file."""
    for filename in ['README.md', 'README.rst', 'README.txt']:
        if os.path.exists(filename):
            with open(filename, encoding='utf-8') as f:
                return f.read()
    return ""

setuptools.setup(
    name='eubi_bridge',
    version='0.0.6',
    author='Bugra Özdemir',
    author_email='bugraa.ozdemir@gmail.com',
    description='A package for converting datasets to OME-Zarr format.',
    long_description=readme(),
    long_description_content_type="text/markdown",
    url='https://github.com/Euro-BioImaging/EuBI-Bridge',
    license='MIT',
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=get_requirements(),
    python_requires='>=3.11,<3.13',
    extras_require={
        # Include CUDA variants if needed
        'cuda11': ['cupy-cuda11x'],
        'cuda12': ['cupy-cuda12x'],
    },
    entry_points={
        'console_scripts': [
            "eubi = eubi_bridge.cli:main"
        ]
    },
)



import shutil, ctypes, time, os, zarr, pprint, psutil, dask, copy
import numpy as np, os, glob, tempfile
from eubi_bridge.base.data_manager import BatchManager
from multiprocessing.pool import ThreadPool

from dask import array as da
from distributed import LocalCluster, Client
from dask_jobqueue import SLURMCluster
from pathlib import Path
from typing import Union

# from eubi_bridge.ngff.multiscales import Pyramid
# from eubi_bridge.ngff import defaults
from eubi_bridge.ebridge_base import BridgeBase, downscale
from eubi_bridge.utils.convenience import sensitive_glob, take_filepaths, is_zarr_group
from eubi_bridge.utils.metadata_utils import print_printable, get_printables

import logging, warnings

logging.getLogger('distributed.diskutils').setLevel(logging.CRITICAL)

def verify_filepaths_for_cluster(
                             filepaths
                             ):
    print(f"Verifying file extensions for distributed setup.")
    formats = ['lif', 'czi', 'lsm',
               'nd2', 'ome.tiff', 'ome.tif',
               'tiff', 'tif', 'zarr',
               'png', 'jpg', 'jpeg']
    for filepath in filepaths:
        verified = any(list(map(lambda path, ext: path.endswith(ext), [filepath] * len(formats), formats)))
        if not verified:
            root, ext = os.path.splitext(filepath)
            warnings.warn(f"Distributed execution is not supported for the {ext} format")
            warnings.warn(f"Falling back on multithreading.")
            break
    if verified:
        print(f"File extensions were verified for distributed setup.")
    return verified

class EuBIBridge:
    """
    EuBIBridge is a conversion tool for bioimage datasets, allowing for both unary and aggregative conversion of image
    data collections to OME-Zarr format.

    Attributes:
        config_gr (zarr.Group): Configuration settings stored in a Zarr group.
        config (dict): Dictionary representation of configuration settings for cluster, conversion, and downscaling.
        dask_config (dict): Dictionary representation of configuration settings for dask.distributed.
        root_defaults (dict): Installation defaults of configuration settings for cluster, conversion, and downscaling.
        root_dask_defaults (dict): Installation defaults of configuration settings for dask.distributed.
    """
    def __init__(self,
                 configpath = f"{os.path.expanduser('~')}/.eubi_bridge",
               ):
        """
        Initializes the EuBIBridge class and loads or sets up default configuration.

        Args:
            configpath (str, optional): Path to store configuration settings. Defaults to the home directory.
        """

        root_dask_defaults = {'distributed.adaptive.interval': '1s', 'distributed.adaptive.maximum': '.inf',
             'distributed.adaptive.minimum': 0, 'distributed.adaptive.target-duration': '5s',
             'distributed.adaptive.wait-count': 3, 'distributed.admin.event-loop': 'tornado',
             'distributed.admin.large-graph-warning-threshold': '10MB',
             'distributed.admin.log-format': '%(asctime)s - %(name)s - %(levelname)s - %(message)s',
             'distributed.admin.log-length': 10000, 'distributed.admin.low-level-log-length': 1000,
             'distributed.admin.max-error-length': 10000, 'distributed.admin.pdb-on-err': False,
             'distributed.admin.system-monitor.disk': True, 'distributed.admin.system-monitor.gil.enabled': True,
             'distributed.admin.system-monitor.gil.interval': '1ms', 'distributed.admin.system-monitor.host-cpu': False,
             'distributed.admin.system-monitor.interval': '500ms', 'distributed.admin.system-monitor.log-length': 7200,
             'distributed.admin.tick.cycle': '1s', 'distributed.admin.tick.interval': '20ms',
             'distributed.admin.tick.limit': '3s', 'distributed.client.heartbeat': '5s',
             'distributed.client.preload': [], 'distributed.client.preload-argv': [],
             'distributed.client.scheduler-info-interval': '2s', 'distributed.client.security-loader': None,
             'distributed.comm.compression': False, 'distributed.comm.default-scheme': 'tcp',
             'distributed.comm.offload': '10MiB', 'distributed.comm.require-encryption': None,
             'distributed.comm.retry.count': 0, 'distributed.comm.retry.delay.max': '20s',
             'distributed.comm.retry.delay.min': '1s', 'distributed.comm.shard': '64MiB',
             'distributed.comm.socket-backlog': 2048, 'distributed.comm.timeouts.connect': '30s',
             'distributed.comm.timeouts.tcp': '30s', 'distributed.comm.tls.ca-file': None,
             'distributed.comm.tls.ciphers': None, 'distributed.comm.tls.client.cert': None,
             'distributed.comm.tls.client.key': None, 'distributed.comm.tls.max-version': None,
             'distributed.comm.tls.min-version': 1.2, 'distributed.comm.tls.scheduler.cert': None,
             'distributed.comm.tls.scheduler.key': None, 'distributed.comm.tls.worker.cert': None,
             'distributed.comm.tls.worker.key': None, 'distributed.comm.ucx.create-cuda-context': None,
             'distributed.comm.ucx.cuda-copy': None, 'distributed.comm.ucx.environment': {},
             'distributed.comm.ucx.infiniband': None, 'distributed.comm.ucx.nvlink': None,
             'distributed.comm.ucx.rdmacm': None, 'distributed.comm.ucx.tcp': None,
             'distributed.comm.websockets.shard': '8MiB', 'distributed.comm.zstd.level': 3,
             'distributed.comm.zstd.threads': 0, 'distributed.dashboard.export-tool': False,
             'distributed.dashboard.graph-max-items': 5000,
             'distributed.dashboard.link': '{scheme}://{host}:{port}/status',
             'distributed.dashboard.prometheus.namespace': 'dask', 'distributed.deploy.cluster-repr-interval': '500ms',
             'distributed.deploy.lost-worker-timeout': '15s',
             'distributed.diagnostics.computations.ignore-files': [r'runpy\.py', 'pytest', r'py\.test',
                                                                   r'pytest-script\.py', '_pytest', 'pycharm',
                                                                   'vscode_pytest', r'get_output_via_markers\.py'],
             'distributed.diagnostics.computations.ignore-modules': ['asyncio', 'functools', 'threading', 'datashader',
                                                                     'dask', 'debugpy', 'distributed', 'coiled', 'cudf',
                                                                     'cuml', 'matplotlib', 'pluggy', 'prefect',
                                                                     'rechunker', 'xarray', 'xgboost', 'xdist',
                                                                     '__channelexec__', 'execnet'],
             'distributed.diagnostics.computations.max-history': 100, 'distributed.diagnostics.computations.nframes': 0,
             'distributed.diagnostics.cudf': False, 'distributed.diagnostics.erred-tasks.max-history': 100,
             'distributed.diagnostics.nvml': True, 'distributed.nanny.environ': {},
             'distributed.nanny.pre-spawn-environ.MALLOC_TRIM_THRESHOLD_': 65536,
             'distributed.nanny.pre-spawn-environ.MKL_NUM_THREADS': 1,
             'distributed.nanny.pre-spawn-environ.OMP_NUM_THREADS': 1,
             'distributed.nanny.pre-spawn-environ.OPENBLAS_NUM_THREADS': 1, 'distributed.nanny.preload': [],
             'distributed.nanny.preload-argv': [], 'distributed.p2p.comm.buffer': '1 GiB',
             'distributed.p2p.comm.concurrency': 10, 'distributed.p2p.comm.message-bytes-limit': '2 MiB',
             'distributed.p2p.comm.retry.count': 10, 'distributed.p2p.comm.retry.delay.max': '30s',
             'distributed.p2p.comm.retry.delay.min': '1s', 'distributed.p2p.storage.buffer': '100 MiB',
             'distributed.p2p.storage.disk': True, 'distributed.p2p.threads': None, 'distributed.rmm.pool-size': None,
             'distributed.scheduler.active-memory-manager.interval': '2s',
             'distributed.scheduler.active-memory-manager.measure': 'optimistic',
             'distributed.scheduler.active-memory-manager.policies': [
                 {'class': 'distributed.active_memory_manager.ReduceReplicas'}],
             'distributed.scheduler.active-memory-manager.start': True, 'distributed.scheduler.allowed-failures': 3,
             'distributed.scheduler.allowed-imports': ['dask', 'distributed'],
             'distributed.scheduler.bandwidth': '100000000', 'distributed.scheduler.blocked-handlers': [],
             'distributed.scheduler.contact-address': None,
             'distributed.scheduler.dashboard.bokeh-application.allow_websocket_origin': ['*'],
             'distributed.scheduler.dashboard.bokeh-application.check_unused_sessions_milliseconds': 500,
             'distributed.scheduler.dashboard.bokeh-application.keep_alive_milliseconds': 500,
             'distributed.scheduler.dashboard.status.task-stream-length': 1000,
             'distributed.scheduler.dashboard.tasks.task-stream-length': 100000,
             'distributed.scheduler.dashboard.tls.ca-file': None, 'distributed.scheduler.dashboard.tls.cert': None,
             'distributed.scheduler.dashboard.tls.key': None, 'distributed.scheduler.default-data-size': '1kiB',
             'distributed.scheduler.default-task-durations.rechunk-split': '1us',
             'distributed.scheduler.default-task-durations.split-shuffle': '1us',
             'distributed.scheduler.default-task-durations.split-stage': '1us',
             'distributed.scheduler.default-task-durations.split-taskshuffle': '1us',
             'distributed.scheduler.events-cleanup-delay': '1h',
             'distributed.scheduler.http.routes': ['distributed.http.scheduler.prometheus',
                                                   'distributed.http.scheduler.info', 'distributed.http.scheduler.json',
                                                   'distributed.http.health', 'distributed.http.proxy',
                                                   'distributed.http.statics'],
             'distributed.scheduler.idle-timeout': None, 'distributed.scheduler.locks.lease-timeout': '30s',
             'distributed.scheduler.locks.lease-validation-interval': '10s',
             'distributed.scheduler.no-workers-timeout': None, 'distributed.scheduler.preload': [],
             'distributed.scheduler.preload-argv': [], 'distributed.scheduler.rootish-taskgroup': 5,
             'distributed.scheduler.rootish-taskgroup-dependencies': 5,
             'distributed.scheduler.unknown-task-duration': '500ms', 'distributed.scheduler.validate': False,
             'distributed.scheduler.work-stealing': True, 'distributed.scheduler.work-stealing-interval': '1s',
             'distributed.scheduler.worker-saturation': 1.1, 'distributed.scheduler.worker-ttl': '5 minutes',
             'distributed.version': 2, 'distributed.worker.blocked-handlers': [],
             'distributed.worker.connections.incoming': 10, 'distributed.worker.connections.outgoing': 50,
             'distributed.worker.daemon': True,
             'distributed.worker.http.routes': ['distributed.http.worker.prometheus', 'distributed.http.health',
                                                'distributed.http.statics'],
             'distributed.worker.lifetime.duration': None, 'distributed.worker.lifetime.restart': False,
             'distributed.worker.lifetime.stagger': '0 seconds', 'distributed.worker.memory.max-spill': False,
             'distributed.worker.memory.monitor-interval': '100ms', 'distributed.worker.memory.pause': 0.8,
             'distributed.worker.memory.rebalance.measure': 'optimistic',
             'distributed.worker.memory.rebalance.recipient-max': 0.6,
             'distributed.worker.memory.rebalance.sender-min': 0.3,
             'distributed.worker.memory.rebalance.sender-recipient-gap': 0.1,
             'distributed.worker.memory.recent-to-old-time': '30s', 'distributed.worker.memory.spill': 0.7,
             'distributed.worker.memory.spill-compression': 'auto', 'distributed.worker.memory.target': 0.6,
             'distributed.worker.memory.terminate': 0.95, 'distributed.worker.memory.transfer': 0.1,
             'distributed.worker.multiprocessing-method': 'spawn', 'distributed.worker.preload': [],
             'distributed.worker.preload-argv': [], 'distributed.worker.profile.cycle': '1000ms',
             'distributed.worker.profile.enabled': True, 'distributed.worker.profile.interval': '10ms',
             'distributed.worker.profile.low-level': False, 'distributed.worker.resources': {},
             'distributed.worker.transfer.message-bytes-limit': '50MB', 'distributed.worker.use-file-locking': True,
             'distributed.worker.validate': False
        }

        defaults = dict(
            cluster = dict(
                n_jobs=4,
                threads_per_worker=1,
                memory_limit='auto',
                temp_dir='auto',
                no_worker_restart=False,
                verbose=False,
                no_distributed=False,
                on_slurm = False,
            ),
            readers = dict(
                as_mosaic=False,
                view_index=0,
                phase_index=0,
                illumination_index=0,
                scene_index=0,
                rotation_index=0,
                mosaic_tile_index=0,
                sample_index=0
            ),                
            conversion = dict(
                zarr_format = 2,
                time_chunk = 1,
                channel_chunk = 1,
                z_chunk = 96,
                y_chunk=96,
                x_chunk=96,
                time_shard_coef = 1,
                channel_shard_coef = 1,
                z_shard_coef = 3,
                y_shard_coef = 3,
                x_shard_coef = 3,
                time_range = None,
                channel_range = None,
                z_range = None,
                y_range = None,
                x_range = None,
                dimension_order='tczyx',
                compressor = 'blosc',
                compressor_params = {},
                overwrite = False,
                use_tensorstore=False,
                use_gpu=False,
                rechunk_method='tasks',
                trim_memory=False,
                metadata_reader = 'bfio',
                save_omexml = True,
                squeeze = False
            ),
            downscale = dict(
                time_scale_factor = 1,
                channel_scale_factor = 1,
                z_scale_factor = 2,
                y_scale_factor = 2,
                x_scale_factor = 2,
                n_layers=3,
                downscale_method='simple',
            )
        )

        self.root_defaults = defaults
        self.root_dask_defaults = root_dask_defaults
        config_gr = zarr.open_group(configpath, mode = 'a')
        config = config_gr.attrs
        for key in defaults.keys():
            if key not in config.keys():
                config[key] = {}
                for subkey in defaults[key].keys():
                    if subkey not in config[key].keys():
                        config[key][subkey] = defaults[key][subkey]
            config_gr.attrs[key] = config[key]
        self.config = dict(config_gr.attrs)
        ###
        if not 'dask_config' in config_gr.keys():
            config_gr.create_group('dask_config')
        dask_config = config_gr['dask_config'].attrs
        for key in root_dask_defaults.keys():
            if key not in dask_config.keys():
                dask_config[key] = root_dask_defaults[key]
        config_gr['dask_config'].attrs.update(dict(dask_config))
        self.dask_config = dict(config_gr['dask_config'].attrs)
        self.config_gr = config_gr
        ###
        self._dask_temp_dir = None
        self.client = None

    def reset_config(self):
        """
        Resets the cluster, conversion and downscale parameters to the installation defaults.
        """
        self.config_gr.attrs.update(self.root_defaults)
        self.config = dict(self.config_gr.attrs)

    def reset_dask_config(self):
        """
        Resets the dask configuration parameters to the installation defaults.
        """
        self.config_gr['dask_config'].attrs.update(self.root_dask_defaults)
        self.dask_config = dict(self.config_gr['dask_config'].attrs)

    def show_config(self):
        """
        Displays the current cluster, conversion, and downscale parameters.
        """
        pprint.pprint(self.config)

    def show_dask_config(self):
        """
        Displays the current dask.distributed parameters.
        """
        pprint.pprint(self.dask_config)

    def show_root_defaults(self):
        """
        Displays the installation defaults for cluster, conversion, and downscale parameters.
        """
        pprint.pprint(self.root_defaults)

    def show_root_dask_defaults(self):
        """
        Displays the installation defaults for dask.distributed.
        """
        pprint.pprint(self.root_dask_defaults)

    def _collect_params(self, param_type, **kwargs):
        """
        Gathers parameters from the configuration, allowing for overrides.

        Args:
            param_type (str): The type of parameters to collect (e.g., 'cluster', 'conversion', 'downscale').
            **kwargs: Parameter values that may override defaults.

        Returns:
            dict: Collected parameters.
        """
        params = {}
        for key in self.config[param_type].keys():
            if key in kwargs.keys():
                params[key] = kwargs[key]
            else:
                params[key] = self.config[param_type][key]
        return params

    def configure_cluster(self,
                          memory_limit: str = 'default',
                          n_jobs: int = 'default',
                          no_worker_restart: bool = 'default',
                          on_slurm: bool = 'default',
                          temp_dir: str = 'default',
                          threads_per_worker: int = 'default',
                          no_distributed: bool = 'default',
                          verbose: bool = 'default'
                         ):
        """
        Updates cluster configuration settings. To update the current default value for a parameter, provide that parameter with a value other than 'default'.

        The following parameters can be configured:
            - memory_limit (str, optional): Memory limit per worker.
            - n_jobs (int, optional): Number of parallel jobs.
            - no_worker_restart (bool, optional): Whether to prevent worker restarts.
            - on_slurm (bool, optional): Whether running on a SLURM cluster.
            - temp_dir (str, optional): Temporary directory for Dask workers.
            - threads_per_worker (int, optional): Number of threads per worker.
            - verbose (bool, optional): Enables detailed logging.

        Args:
            memory_limit (str, optional): Memory limit per worker.
            n_jobs (int, optional): Number of parallel jobs.
            no_worker_restart (bool, optional): Whether to prevent worker restarts.
            on_slurm (bool, optional): Whether running on a SLURM cluster.
            temp_dir (str, optional): Temporary directory for Dask workers.
            threads_per_worker (int, optional): Number of threads per worker.
            verbose (bool, optional): Enables detailed logging.

        Returns:
            None
        """

        params = {
            'memory_limit': memory_limit,
            'n_jobs': n_jobs,
            'no_worker_restart': no_worker_restart,
            'on_slurm': on_slurm,
            'temp_dir': temp_dir,
            'threads_per_worker': threads_per_worker,
            'no_distributed': no_distributed,
            'verbose': verbose
        }

        for key in params:
            if key in self.config['cluster'].keys():
                if params[key] != 'default':
                    self.config['cluster'][key] = params[key]
        self.config_gr.attrs['cluster'] = self.config['cluster']

    def configure_readers(self,
                          as_mosaic: bool = 'default',
                          view_index: int = 'default',
                          phase_index: int = 'default',
                          illumination_index: int = 'default',
                          scene_index: int = 'default',
                          rotation_index: int = 'default',
                          mosaic_tile_index: int = 'default',
                          sample_index: int = 'default',
                         ):
        """
        Updates reader configuration settings. To update the current default value for a parameter, provide that parameter with a value other than 'default'.

        Returns:
            None
        """

        params = {
            'as_mosaic': as_mosaic,
            'view_index': view_index,
            'phase_index': phase_index,
            'illumination_index': illumination_index,
            'scene_index': scene_index,
            'rotation_index': rotation_index,
            'mosaic_tile_index': mosaic_tile_index,
            'sample_index': sample_index
        }

        for key in params:
            if key in self.config['readers'].keys():
                if params[key] != 'default':
                    self.config['readers'][key] = params[key]
        self.config_gr.attrs['readers'] = self.config['readers']


    def configure_conversion(self,
                             zarr_format: int = 'default',
                             compressor: str = 'default',
                             compressor_params: dict = 'default',
                             time_chunk: int = 'default',
                             channel_chunk: int = 'default',
                             z_chunk: int = 'default',
                             y_chunk: int = 'default',
                             x_chunk: int = 'default',
                             time_shard_coef: int = 'default',
                             channel_shard_coef: int = 'default',
                             z_shard_coef: int = 'default',
                             y_shard_coef: int = 'default',
                             x_shard_coef: int = 'default',
                             time_range: int = 'default',
                             channel_range: int = 'default',
                             z_range: int = 'default',
                             y_range: int = 'default',
                             x_range: int = 'default',
                             dimension_order: str = 'default',
                             overwrite: bool = 'default',
                             rechunk_method: str = 'default',
                             # rechunkers_max_mem: str = 'default',
                             trim_memory: bool = 'default',
                             use_tensorstore: bool = 'default',
                             use_gpu: bool = 'default',
                             metadata_reader: str = 'default',
                             save_omexml: bool = 'default',
                             squeeze: bool = 'default'
                             ):
        """
        Updates conversion configuration settings. To update the current default value for a parameter, provide that parameter with a value other than 'default'.

        The following parameters can be configured:
            - compressor (str, optional): Compression algorithm.
            - compressor_params (dict, optional): Parameters for the compressor.
            - output_chunks (list, optional): Chunk size for output.
            - overwrite (bool, optional): Whether to overwrite existing data.
            - rechunk_method (str, optional): Method used for rechunking.
            - trim_memory (bool, optional): Whether to trim memory usage.
            - use_tensorstore (bool, optional): Whether to use TensorStore for writing.
            - save_omexml (bool, optional): Whether to create a METADATA.ome.xml file.
        Args:
            compressor (str, optional): Compression algorithm.
            compressor_params (dict, optional): Parameters for the compressor.
            output_chunks (list, optional): Chunk size for output.
            overwrite (bool, optional): Whether to overwrite existing data.
            rechunk_method (str, optional): Method used for rechunking.
            trim_memory (bool, optional): Whether to trim memory usage.
            use_tensorstore (bool, optional): Whether to use TensorStore for storage.
            save_omexml (bool, optional): Whether to create a METADATA.ome.xml file.

        Returns:
            None
        """

        params = {
            'zarr_format': zarr_format,
            'compressor': compressor,
            'compressor_params': compressor_params,
            "time_chunk": time_chunk,
            "channel_chunk": channel_chunk,
            "z_chunk": z_chunk,
            "y_chunk": y_chunk,
            "x_chunk": x_chunk,
            "time_shard_coef": time_shard_coef,
            "channel_shard_coef": channel_shard_coef,
            "z_shard_coef": z_shard_coef,
            "y_shard_coef": y_shard_coef,
            "x_shard_coef": x_shard_coef,
            "time_range": time_range,
            "channel_range": channel_range,
            "z_range": z_range,
            "y_range": y_range,
            "x_range": x_range,
            "dimension_order": dimension_order,
            'overwrite': overwrite,
            'rechunk_method': rechunk_method,
            # 'rechunkers_max_mem': rechunkers_max_mem,
            'trim_memory': trim_memory,
            'use_tensorstore': use_tensorstore,
            'use_gpu': use_gpu,
            'metadata_reader': metadata_reader,
            'save_omexml': save_omexml,
            'squeeze': squeeze
        }

        for key in params:
            if key in self.config['conversion'].keys():
                if params[key] != 'default':
                    self.config['conversion'][key] = params[key]
        self.config_gr.attrs['conversion'] = self.config['conversion']

    def configure_downscale(self,
                            downscale_method: str = 'default',
                            n_layers: int = 'default',
                            time_scale_factor: int = 'default',
                            channel_scale_factor: int = 'default',
                            z_scale_factor: int = 'default',
                            y_scale_factor: int = 'default',
                            x_scale_factor: int = 'default',
                            ):
        """
        Updates downscaling configuration settings. To update the current default value for a parameter, provide that parameter with a value other than 'default'.

        The following parameters can be configured:
            - downscale_method (str, optional): Downscaling algorithm.
            - n_layers (int, optional): Number of downscaling layers.
            - scale_factor (list, optional): Scaling factors for each dimension.

        Args:
            downscale_method (str, optional): Downscaling algorithm.
            n_layers (int, optional): Number of downscaling layers.
            scale_factor (list, optional): Scaling factors for each dimension.

        Returns:
            None
        """

        params = {
            'downscale_method': downscale_method,
            'n_layers': n_layers,
            'time_scale_factor': time_scale_factor,
            "channel_scale_factor": channel_scale_factor,
            "z_scale_factor": z_scale_factor,
            "y_scale_factor": y_scale_factor,
            "x_scale_factor": x_scale_factor,
        }

        for key in params:
            if key in self.config['downscale'].keys():
                if params[key] != 'default':
                    self.config['downscale'][key] = params[key]
        self.config_gr.attrs['downscale'] = self.config['downscale']

    def _set_dask_temp_dir(self, temp_dir = 'auto'):
        if self._dask_temp_dir is not None:
            self._dask_temp_dir.cleanup()
        if temp_dir in ('auto', None):
            temp_dir = tempfile.TemporaryDirectory()
        else:
            os.makedirs(temp_dir, exist_ok=True)
            temp_dir = tempfile.TemporaryDirectory(dir=temp_dir)
        self._dask_temp_dir = temp_dir
        return self

    def _start_cluster(self,
                    n_jobs: int = 4,
                    threads_per_worker: int = 1,
                    memory_limit: str = 'auto',
                    temp_dir='auto',
                    verbose = False,
                    on_slurm = False,
                    no_distributed = False,
                    config_kwargs = {},
                    **kwargs
                    ):

        config_dict = copy.deepcopy(self.dask_config)
        config_dict.update(**config_kwargs)

        # worker_options = {
        #     "memory_target_fraction": 0.8,
        #     "memory_spill_fraction": 0.9,
        #     "memory_pause_fraction": 0.95,
        #     # "memory_terminate_fraction": 0.98
        # }
        scheduler_options = {
            "allowed_failures": 100,
            "idle_timeout": "1h",
            "worker_ttl": "1d"  # Set to a large value, e.g., 1 day
        }

        self._set_dask_temp_dir(temp_dir)

        dask.config.set(config_dict) # use dictionary notation here.

        if no_distributed:
            config_dict.update(scheduler = 'threads',
                               pool = ThreadPool(n_jobs)
                               )
            dask.config.set(config_dict)
            print(f"Process running locally via multithreading.")
        else:
            if memory_limit == 'auto':
                reserve_fraction = kwargs.get('reserve_memory_fraction', 0.1)
                min_per_worker = kwargs.get('min_memory_per_worker', 1 * 1024 ** 3)

                total_mem = psutil.virtual_memory().total
                reserved_mem = total_mem * reserve_fraction
                available_mem = max(0, total_mem - reserved_mem)
                mem_per_worker = max(available_mem / n_jobs, min_per_worker)
                mem_gb = mem_per_worker / (1 * 1024 ** 3)
                memory_limit = f"{mem_gb} GB"
                print(f"{memory_limit} memory has been allocated per worker.")

            if on_slurm:
                print(f"Process running on Slurm.")
                cluster = SLURMCluster(
                                        cores=threads_per_worker,
                                        processes=1,
                                        nanny=False,
                                        scheduler_options=scheduler_options,
                                        n_workers=n_jobs,
                                        memory=memory_limit,
                                        local_directory=f"{self._dask_temp_dir.name}",
                                        # **worker_options
                                        )
            else:
                print(f"Process running on local cluster.")
                cluster = LocalCluster(
                                       n_workers=n_jobs,
                                       threads_per_worker=threads_per_worker,
                                       nanny = False,
                                       scheduler_kwargs=scheduler_options,
                                       memory_limit = memory_limit,
                                       local_directory=f"{self._dask_temp_dir.name}",
                                       # **worker_options
                                       )
            cluster.scale(n_jobs)
            self.client = Client(cluster)
            if verbose:
                print(self.client.cluster)
        return self

    def to_zarr(self,
                input_path: Union[Path, str],
                output_path: Union[Path, str],
                includes=None,
                excludes=None,
                time_tag: Union[str, tuple] = None,
                channel_tag: Union[str, tuple] = None,
                z_tag: Union[str, tuple] = None,
                y_tag: Union[str, tuple] = None,
                x_tag: Union[str, tuple] = None,
                concatenation_axes: Union[int, tuple, str] = None,
                **kwargs
                ):
        """
        Converts image data to OME-Zarr format and optionally applies downscaling.

        Args:
            input_path (Union[Path, str]): Path to input file or directory.
            output_path (Union[Path, str]): Directory, in which the output OME-Zarrs will be written.
            includes (str, optional): Filename patterns to filter for.
            excludes (str, optional): Filename patterns to filter against.
            time_tag (Union[str, tuple], optional): Time dimension tag.
            channel_tag (Union[str, tuple], optional): Channel dimension tag.
            z_tag (Union[str, tuple], optional): Z dimension tag.
            y_tag (Union[str, tuple], optional): Y dimension tag.
            x_tag (Union[str, tuple], optional): X dimension tag.
            concatenation_axes (Union[int, tuple, str], optional): Axes, along which the images will be concatenated.
            **kwargs: Additional configuration overrides.

        Raises:
            Exception: If no files are found in the input path.

        Prints:
            Process logs including conversion and downscaling time.

        Returns:
            None
        """
        t0 = time.time()
        # Get parameters:
        self.cluster_params = self._collect_params('cluster', **kwargs)
        self.readers_params = self._collect_params('readers', **kwargs)
        self.conversion_params = self._collect_params('conversion', **kwargs)
        self.downscale_params = self._collect_params('downscale', **kwargs)

        if self.conversion_params['use_gpu'] and self.conversion_params['use_tensorstore']:
            raise ValueError("Tensorstore is not supported for GPU arrays.")

        print(f"Base conversion initiated.")
        ###### Handle input data and metadata
        paths = take_filepaths(input_path, includes = includes, excludes = excludes)

        filepaths = sorted(list(paths))

        ###### Start the cluster
        verified_for_cluster = verify_filepaths_for_cluster(filepaths) ### Ensure non-bioformats conversion. If bioformats is needed, fall back on local conversion.
        if not verified_for_cluster:
            self.cluster_params['no_distributed'] = True

        cluster_is_true = not self.cluster_params['no_distributed']

        if cluster_is_true:
            chunks_yx = None
        else:
            chunks_yx = tuple([self.conversion_params['y_chunk'],  # chunks already when reading
                               self.conversion_params['x_chunk']]
                              )

        self._start_cluster(**self.cluster_params)


        series = self.readers_params['scene_index']

        ###### Read and concatenate
        base = BridgeBase(input_path,
                        excludes=excludes,
                        includes=includes,
                        series=series,
                        zarr_format = self.conversion_params['zarr_format'],
                        verbose = self.cluster_params['verbose']
                        )

        base.read_dataset(cluster_is_true,
                          chunks_yx=chunks_yx,
                          readers_params = self.readers_params
                          )

        base.digest(
            time_tag = time_tag,
            channel_tag = channel_tag,
            z_tag = z_tag,
            y_tag = y_tag,
            x_tag = x_tag,
            axes_of_concatenation = concatenation_axes,
            metadata_reader = self.conversion_params['metadata_reader'],
            **kwargs
            )
        print(f"Metadata was extracted")
        verbose = base._verbose

        if 'region_shape' in kwargs.keys():
            self.conversion_params['region_shape'] = kwargs.get('region_shape')
        if verbose:
            print(f"Cluster params:")
            pprint.pprint(self.cluster_params)
            print(f"Readers params:")
            pprint.pprint(self.readers_params)
            print(f"Conversion params:")
            pprint.pprint(self.conversion_params)
            print(f"Downscale params:")
            pprint.pprint(self.downscale_params)

        temp_dir = base._dask_temp_dir
        self.conversion_params['temp_dir'] = temp_dir
        self.downscale_params['temp_dir'] = temp_dir

        if self.client is not None:
            base.client = self.client
        base.set_dask_temp_dir(self._dask_temp_dir)

        ###### Write
        self.base_results = base.write_arrays(output_path,
                                           compute=True,
                                           verbose=verbose,
                                           **self.conversion_params
                                           )
        ###### Downscale
        print(f"Base conversion finished.")
        t1 = time.time()
        print(f"Elapsed for base conversion: {(t1 - t0) / 60} min.")
        n_layers = self.downscale_params['n_layers']
        if n_layers > 1:
            print(f"Downscaling initiated.")
            _ = downscale(
                      self.base_results,
                      **self.downscale_params,
                      zarr_format = self.conversion_params['zarr_format'],
                      rechunk_method = self.conversion_params['rechunk_method'],
                      use_tensorstore = self.conversion_params['use_tensorstore'],
                      verbose = verbose
                      ) # TODO: add to_cupy parameter here.

            print(f"Downscaling finished.")

        ###### Shutdown and clean up
        if self.client is not None:
            self.client.shutdown()
            self.client.close()

        if isinstance(self._dask_temp_dir, tempfile.TemporaryDirectory):
            shutil.rmtree(self._dask_temp_dir.name)
        else:
            shutil.rmtree(self._dask_temp_dir)

        t1 = time.time()
        print(f"Elapsed for conversion + downscaling: {(t1 - t0) / 60} min.")


    def show_pixel_meta(self,
                input_path: Union[Path, str],
                includes=None,
                excludes=None,
                series: int = None, # self.readers_params['scene_index'],
                **kwargs
                ):
        """
        Print pixel-level metadata for all datasets in the 'input_path'.

        Args:
            input_path (Union[Path, str]): Path to input file or directory.
            output_path (Union[Path, str]): Directory, in which the output OME-Zarrs will be written.
            includes (str, optional): Filename patterns to filter for.
            excludes (str, optional): Filename patterns to filter against.
            **kwargs: Additional configuration overrides.

        Raises:
            Exception: If no files are found in the input path.

        Prints:
            Process logs including conversion and downscaling time.

        Returns:
            None
        """

        # Get parameters:
        self.cluster_params = self._collect_params('cluster', **kwargs)
        self.readers_params = self._collect_params('readers', **kwargs)
        self.conversion_params = self._collect_params('conversion', **kwargs)

        paths = take_filepaths(input_path, includes = includes, excludes = excludes)

        filepaths = sorted(list(paths))

        ###### Start the cluster
        verified_for_cluster = verify_filepaths_for_cluster(filepaths)
        if not verified_for_cluster:
            self.cluster_params['no_distributed'] = True

        self._start_cluster(**self.cluster_params)

        series = self.readers_params['scene_index']

        ###### Read and digest
        base = BridgeBase(input_path,
                        excludes=excludes,
                        includes=includes,
                        series=series
                        )

        base.read_dataset(verified_for_cluster,
                          readers_params=self.readers_params
                          )

        base.digest()

        temp_dir = base._dask_temp_dir
        self.conversion_params['temp_dir'] = temp_dir

        if self.client is not None:
            base.client = self.client
        base.set_dask_temp_dir(self._dask_temp_dir)

        printables = {
            path: get_printables(
                manager.axes,
                manager.shapedict,
                manager.scaledict,
                manager.unitdict
            )
            for path, manager in base.batchdata.managers.items()
        }
        for path, printable in printables.items():
            print('---------')
            print(f"")
            print(f"Metadata for '{path}':")
            print_printable(printable)

        ###### Shutdown and clean up
        if self.client is not None:
            self.client.shutdown()
            self.client.close()

        if isinstance(self._dask_temp_dir, tempfile.TemporaryDirectory):
            shutil.rmtree(self._dask_temp_dir.name)
        else:
            shutil.rmtree(self._dask_temp_dir)

    def update_pixel_meta(self,
                          input_path: Union[Path, str],
                          includes=None,
                          excludes=None,
                          time_scale: (int, float) = None,
                          z_scale: (int, float) = None,
                          y_scale: (int, float) = None,
                          x_scale: (int, float) = None,
                          time_unit: str = None,
                          z_unit: str = None,
                          y_unit: str = None,
                          x_unit: str = None,
                          **kwargs
                          ):
        """
        Updates pixel metadata for image files located at the specified input path.

        Args:
            input_path (Union[Path, str]): Path to input file or directory.
            includes (optional): Filename patterns to include.
            excludes (optional): Filename patterns to exclude.
            series (int, optional): Series index to process.
            time_scale, z_scale, y_scale, x_scale ((int, float), optional): Scaling factors for the respective dimensions.
            time_unit, z_unit, y_unit, x_unit (str, optional): Units for the respective dimensions.
            **kwargs: Additional parameters for cluster and conversion configuration.

        Returns:
            None
        """

        # Collect cluster and conversion parameters
        self.cluster_params = self._collect_params('cluster', **kwargs)
        self.readers_params = self._collect_params('readers', **kwargs)
        self.conversion_params = self._collect_params('conversion', **kwargs)

        # Collect file paths based on inclusion and exclusion patterns
        paths = take_filepaths(input_path, includes=includes, excludes=excludes)

        filepaths = sorted(list(paths))

        # Verify file paths for cluster compatibility
        verified_for_cluster = verify_filepaths_for_cluster(filepaths)
        if not verified_for_cluster:
            self.cluster_params['no_distributed'] = True

        cluster_is_true = not self.cluster_params['no_distributed']

        # Start the processing cluster
        self._start_cluster(**self.cluster_params)

        series = self.readers_params['scene_index']

        base = BridgeBase(input_path,
                          excludes=excludes,
                          includes=includes,
                          series=series
                          )

        # Read and digest the dataset
        base.read_dataset(verified_for_cluster,
                          readers_params=self.readers_params
                          )

        # Prepare pixel metadata arguments
        pixel_meta_kwargs_ = dict(time_scale=time_scale,
                                  z_scale=z_scale,
                                  y_scale=y_scale,
                                  x_scale=x_scale,
                                  time_unit=time_unit,
                                  z_unit=z_unit,
                                  y_unit=y_unit,
                                  x_unit=x_unit)
        pixel_meta_kwargs = {key: val for key, val in pixel_meta_kwargs_.items() if val is not None}

        base.digest(**pixel_meta_kwargs)
        print(f"Metadata was extracted")

        if self.client is not None:
            base.client = self.client
        base.set_dask_temp_dir(self._dask_temp_dir)

        # Update metadata for each dataset manager
        for path, manager in base.batchdata.managers.items():
            if is_zarr_group(manager.path):
                manager.sync_pyramid(self.conversion_params['save_omexml'])
            else:
                print(f"Cannot update metadata for non-zarr path: {path}")

        # Shutdown the cluster and clean up temporary directories
        if self.client is not None:
            self.client.shutdown()
            self.client.close()

        if isinstance(self._dask_temp_dir, tempfile.TemporaryDirectory):
            shutil.rmtree(self._dask_temp_dir.name)
        else:
            shutil.rmtree(self._dask_temp_dir)






import os, json, glob, time
import warnings

import numpy as np, pandas as pd
try:
    import cupy as cp
    cupy_available = True
except:
    cupy_available = False
import zarr, json, shutil, os, copy, zarr
from dask import array as da
import dask
import numcodecs
import shutil, tempfile
from pathlib import Path
from typing import List

from typing import (
    Union,
    Tuple,
    Dict,
    Any,
    Iterable,
    List,
    Optional
)

def asstr(s):
    if isinstance(s, str):
        return s
    elif isinstance(s, int):
        return str(s)
    else:
        raise TypeError(f"Input must be either of types {str, int}")

def asdask(data, chunks = 'auto'):
    assert isinstance(data, (da.Array, zarr.Array, np.ndarray)), f'data must be of type: {da.Array, zarr.Array, np.ndarray}'
    if isinstance(data, zarr.Array):
        return da.from_zarr(data)
    elif isinstance(data, np.ndarray):
        return da.from_array(data, chunks = chunks)
    return data

def path_has_pyramid(path):
    try:
        store = zarr.storage.LocalStore(path)
        _ = zarr.open_group(store, mode = 'r')
        return True
    except:
        return False


def parse_as_list(path_or_paths: Union[Iterable, str, int, float]
                   ):
    if isinstance(path_or_paths, (str, int, float)):
        inputs = [path_or_paths]
    else:
        inputs = path_or_paths
    return inputs

def includes(group1: Union[Iterable, str, int, float],
             group2: Union[Iterable, str, int, float]
             ):
    """Convenience function that checks if group1 includes group2 completely."""
    gr1 = parse_as_list(group1)
    gr2 = parse_as_list(group2)
    return all([item in gr1 for item in gr2])

def insert_at_indices(iterable1, iterable2, indices):
    if not hasattr(iterable1, '__len__'):
        iterable1 = [iterable1]
    if not hasattr(iterable2, '__len__'):
        iterable2 = [iterable2]
    if not hasattr(indices, '__len__'):
        indices = [indices]
    endlen = (len(iterable1) + len(iterable2))
    end_indices = [None] * endlen
    other_indices = [i for i in range(endlen) if i not in indices]
    for i, j in zip(other_indices, iterable1):
        end_indices[i] = j
    for i, j in zip(indices, iterable2):
        end_indices[i] = j
    return end_indices

def index_nth_dimension(array,
                        dimensions = 2, # a scalar or iterable
                        intervals = None # a scalar, an iterable of scalars, a list of tuple or None
                        ):
    if isinstance(array, zarr.Array):
        array = da.from_zarr(array)
    allinds = np.arange(array.ndim).astype(int)
    if np.isscalar(dimensions):
        dimensions = [dimensions]
    if intervals is None or np.isscalar(intervals):
        intervals = np.repeat(intervals, len(dimensions))
    assert len(intervals) == len(dimensions) ### KALDIM
    interval_dict = {item: interval for item, interval in zip(dimensions, intervals)}
    shape = array.shape
    slcs = []
    for idx, dimlen in zip(allinds, shape):
        if idx not in dimensions:
            slc = slice(dimlen)
        else:
            try:
                slc = slice(interval_dict[idx][0], interval_dict[idx][1])
            except:
                slc = interval_dict[idx]
        slcs.append(slc)
    slcs = tuple(slcs)
    indexed = array[slcs]
    return indexed

def transpose_dict(dictionary):
    keys, values = [], []
    for key, value in dictionary.items():
        keys.append(key)
        values.append(value)
    return keys, values

def argsorter(s):
    return sorted(range(len(s)), key = lambda k: s[k])


def is_zarr_array(path: (str, Path)
                  ):
    try:
        _ = zarr.open_array(path, mode = 'r')
        return True
    except:
        return False

######## GROUP UTILITIES BELOW
def is_zarr_group(path: (str, Path)
                  ):
    try:
        _ = zarr.open_group(path, mode = 'r')
        return True
    except:
        return False

def is_generic_collection(group):
    res = False
    basepath = group.store.path
    basename = os.path.basename(basepath)
    paths = list(group.keys())
    attrs = dict(group.attrs)
    attrkeys, attrvalues = transpose_dict(attrs)
    if basename in attrkeys and (len(paths) > 0):
        if len(attrs[basename]) == len(paths):
            res = True
            for item0, item1 in zip(attrs[basename], paths):
                if item0 != item1:
                    res = False
    return res

def get_collection_paths(directory,
                         return_all = False
                         ):
    gr = zarr.group(directory)
    groupkeys = list(gr.group_keys())
    arraykeys = list(gr.array_keys())
    grouppaths = [os.path.join(directory, item) for item in groupkeys]
    arraypaths = [os.path.join(directory, item) for item in arraykeys]
    collection_paths = []
    multiscales_paths = []
    while len(grouppaths) > 0:
        if is_generic_collection(gr) or 'bioformats2raw.layout' in gr.attrs:
            collection_paths.append(directory)
        if 'multiscales' in list(gr.attrs.keys()):
            multiscales_paths.append(directory)
        directory = grouppaths[0]
        grouppaths.pop(0)
        gr = zarr.group(directory)
        groupkeys = list(gr.group_keys())
        arraykeys = list(gr.array_keys())
        grouppaths += [os.path.join(directory, item) for item in groupkeys]
        arraypaths += [os.path.join(directory, item) for item in arraykeys]
    if is_generic_collection(gr) or 'bioformats2raw.layout' in gr.attrs:
        collection_paths.append(directory)
    if 'multiscales' in list(gr.attrs.keys()):
        multiscales_paths.append(directory)
    out = [item for item in collection_paths]
    for mpath in multiscales_paths:
        s = os.path.dirname(mpath)
        if s in collection_paths:
            pass
        else:
            if mpath not in out:
                out.append(mpath)
    if return_all:
        return out, multiscales_paths, arraypaths
    return out

def convert_np_types(obj):
    if isinstance(obj, np.integer):
        return int(obj)
    elif isinstance(obj, np.floating):
        return float(obj)
    elif isinstance(obj, np.ndarray):
        return obj.tolist()
    else:
        return obj

def is_valid_json(my_dict):
    try:
        json.dumps(my_dict)
        return True
    except:
        warnings.warn(f"Object is not a valid json!")
        return False

def turn2json(my_dict):
    stringified = json.dumps(my_dict, default = convert_np_types)
    return json.loads(stringified)


def as_store(store: (zarr.storage.StoreLike, zarr.Array, Path, str) # TODO: other scenarios?
                ):
    assert isinstance(store, (zarr.storage.Store, zarr.Array, Path, str)), f"The given store cannot be parsed."
    if isinstance(store, (Path, str)):
        out = zarr.storage.LocalStore(store, dimension_separator = '/')
    else:
        out = store
    return out

def as_dask_array(array: (da.Array, zarr.Array, np.ndarray,
                          # cp.ndarray
                          ),
                  backend = 'numpy',
                  **params
                  ):
    if cupy_available:
        assert isinstance(array, (da.Array, zarr.Array, np.ndarray, cp.ndarray)), f"The given array type {type(array)} cannot be parsed."
    else:
        assert isinstance(array, (da.Array, zarr.Array, np.ndarray)), f"The given array type {type(array)} cannot be parsed."
    assert backend in ('numpy', 'cupy'), f"Currently, the only supported backends are 'numpy' or 'cupy'."

    if not isinstance(array, da.Array):
        out = da.from_array(array, **params)

    if backend == 'cupy':
        if cupy_available:
            out = out.map_blocks(cp.asarray)
        else:
            raise ValueError("cupy is not available!")

    return out


#########################################################

def get_array_size(array, as_str = True):
    voxelcount = np.prod(array.shape)
    arraysize = voxelcount * array.dtype.itemsize
    if as_str:
        return f"{arraysize / 1024 ** 3}GB"
    else:
        return arraysize

def sizeof(array, unit = 'gb'):
    unit = unit.lower()
    assert unit in ('gb', 'mb', 'kb')
    bytes = get_array_size(array, False)
    if unit == 'gb':
        ret = bytes / 1024 ** 3
    elif unit == 'mb':
        ret = bytes / 1024 ** 2
    elif unit == 'kb':
        ret = bytes / 1024 ** 1
    else:
        ret = bytes
    return ret

def get_chunksize_from_array(arr):
    chunks = arr.chunksize if isinstance(arr, da.Array) else arr.chunks
    itemsize = arr.dtype.itemsize
    chunk_size = itemsize * np.prod(chunks)
    chunk_size_gb = chunk_size / (1024 ** 3)
    return f'{chunk_size_gb * 1.1}GB'


def retry_decorator(retries=3, delay=1, exceptions=(Exception,)):
    def decorator(func):
        def wrapper(*args, **kwargs):
            for attempt in range(retries):
                try:
                    return func(*args, **kwargs)
                except exceptions:
                    if attempt < retries - 1:
                        time.sleep(delay)
                    else:
                        raise
        return wrapper
    return decorator



def sensitive_glob(pattern: str,
                   recursive: bool = False,
                   sensitive_to: str = '.zarr'
                   ) -> List[str]:
    results = []

    for start_path in glob.glob(pattern, recursive=recursive):
        def _walk(current_path):
            if os.path.isfile(current_path):
                results.append(current_path)
                return
            if os.path.isdir(current_path):
                if current_path.endswith(sensitive_to):
                    results.append(current_path)
                    return
                for entry in os.listdir(current_path):
                    entry_path = os.path.join(current_path, entry)
                    _walk(entry_path)

        _walk(start_path)

    return results


def take_filepaths(input_path: str,
                   includes: bool = None,
                   excludes: bool = None
                   ):

    if os.path.isfile(input_path) or input_path.endswith('.zarr'):
        dirname = os.path.dirname(input_path)
        basename = os.path.basename(input_path)
        input_path = f"{dirname}/*{basename}"

    if not '*' in input_path and not input_path.endswith('.zarr'):
        input_path = os.path.join(input_path, '**')

    # print(input_path)
    if not '*' in input_path:
        input_path_ = os.path.join(input_path, '**')
    else:
        input_path_ = input_path
    paths = sensitive_glob(input_path_, recursive=False, sensitive_to='.zarr')

    paths = list(filter(lambda path: (includes in path if includes is not None else True)
                                     and
                                     (excludes not in path if excludes is not None else True),
                        paths
                        )
                 )
    paths = list(filter(lambda path: not path.endswith('zarr.json'), paths))
    return sorted(paths)

